@@ -19,6 +19,11 @@ class _PyRunnerConfig(_RunnerConfig):
     name = "py"
 
 
+@dataclasses.dataclass(frozen=True)
+class _DynamicRunnerConfig(_RunnerConfig):
+    name = "dynamic"
+
+
 @dataclasses.dataclass(frozen=True)
 class _RayRunnerConfig(_RunnerConfig):
     name = "ray"
@@ -39,6 +44,8 @@ def _get_runner_config_from_env() -> _RunnerConfig:
             return _RayRunnerConfig(address=os.getenv("DAFT_RAY_ADDRESS"))
         elif runner.upper() == "PY":
             return _PyRunnerConfig()
+        elif runner.upper() == "DYNAMIC":
+            return _DynamicRunnerConfig()
         raise ValueError(f"Unsupported DAFT_RUNNER variable: {os.environ['DAFT_RUNNER']}")
     return _PyRunnerConfig()
 
@@ -69,6 +76,12 @@ class DaftContext:
 
             logger.info("Using PyRunner")
             _RUNNER = PyRunner()
+
+        elif self.runner_config.name == "dynamic":
+            from daft.runners.dynamic_runner import DynamicRunner
+
+            logger.info("Using DynamicRunner")
+            _RUNNER = DynamicRunner()
         else:
             raise NotImplementedError(f"Runner config implemented: {self.runner_config.name}")
 
@@ -132,3 +145,15 @@ def set_runner_py() -> DaftContext:
         disallow_set_runner=True,
     )
     return _DaftContext
+
+
+def set_runner_dynamic() -> DaftContext:
+    global _DaftContext
+    if _DaftContext.disallow_set_runner:
+        raise RuntimeError("Cannot set runner more than once")
+    _DaftContext = dataclasses.replace(
+        _DaftContext,
+        runner_config=_DynamicRunnerConfig(),
+        disallow_set_runner=True,
+    )
+    return _DaftContext
