@@ -1,32 +1,37 @@
 from __future__ import annotations
 
-import io
+import warnings
 from dataclasses import dataclass
-from functools import partial
-from typing import IO, TYPE_CHECKING, Any, Callable, Iterable, TypeVar, Union
+from typing import TYPE_CHECKING, Any, Callable, Iterable, TypeVar, Union
 
 import pandas
 import pyarrow as pa
-import pyarrow.parquet as papq
-from pyarrow import csv, json
 
 from daft.context import get_context
+from daft.dataframe.preview import DataFramePreview
 from daft.dataframe.schema import DataFrameSchema
 from daft.datasources import (
     CSVSourceInfo,
     JSONSourceInfo,
     ParquetSourceInfo,
+    SourceInfo,
     StorageType,
 )
 from daft.errors import ExpressionTypeError
 from daft.execution.operators import ExpressionType
 from daft.expressions import ColumnExpression, Expression, col
-from daft.filesystem import get_filesystem_from_path, get_protocol_from_path
+from daft.filesystem import get_filesystem_from_path
 from daft.logical import logical_plan
 from daft.logical.schema import ExpressionList
-from daft.runners.partitioning import PartitionCacheEntry, PartitionSet, vPartition
+from daft.runners.partitioning import (
+    PartitionCacheEntry,
+    PartitionSet,
+    vPartition,
+    vPartitionParseCSVOptions,
+    vPartitionReadOptions,
+    vPartitionSchemaInferenceOptions,
+)
 from daft.runners.pyrunner import LocalPartitionSet
-from daft.runners.ray_runner import RayPartitionSet
 from daft.types import PythonExpressionType
 from daft.viz import DataFrameDisplay
 
@@ -38,35 +43,43 @@ UDFReturnType = TypeVar("UDFReturnType", covariant=True)
 ColumnInputType = Union[Expression, str]
 
 
-def _sample_with_pyarrow(
-    loader_func: Callable[[IO], pa.Table],
-    filepath: str,
-    max_bytes: int = 5 * 1024**2,
-) -> ExpressionList:
-    fs = get_filesystem_from_path(filepath)
-    sampled_bytes = io.BytesIO()
-    with fs.open(filepath, compression="infer") as f:
-        lines = f.readlines(max_bytes)
-        for line in lines:
-            sampled_bytes.write(line)
-    sampled_bytes.seek(0)
-    sampled_tbl = loader_func(sampled_bytes)
-    fields = [(field.name, field.type) for field in sampled_tbl.schema]
-    schema = ExpressionList(
-        [ColumnExpression(name, expr_type=ExpressionType.from_arrow_type(type_)) for name, type_ in fields]
+def _get_tabular_files_scan(
+    path: str, get_schema: Callable[[str], ExpressionList], source_info: SourceInfo
+) -> logical_plan.TabularFilesScan:
+    """Returns a TabularFilesScan LogicalPlan for a given glob filepath."""
+    # Glob the path and return as a DataFrame with a column containing the filepaths
+    partition_set_factory = get_context().runner().partition_set_factory()
+    partition_set, filepaths_schema = partition_set_factory.glob_filepaths(path)
+    cache_entry = get_context().runner().put_partition_set_into_cache(partition_set)
+    filepath_plan = logical_plan.InMemoryScan(
+        cache_entry=cache_entry,
+        schema=filepaths_schema,
+        partition_spec=logical_plan.PartitionSpec(logical_plan.PartitionScheme.UNKNOWN, partition_set.num_partitions()),
     )
-    assert schema is not None, f"Unable to read file {filepath} to determine schema"
-    return schema
+    filepath_df = DataFrame(filepath_plan)
 
+    # Sample the first 10 filepaths and infer the schema
+    schema_df = filepath_df.limit(10).select(
+        col(partition_set_factory.FILEPATH_COLUMN_NAME).apply(get_schema, return_type=ExpressionList).alias("schema")
+    )
+    schema_df.collect()
+    schema_result = schema_df._result
+    assert schema_result is not None
+    sampled_schemas = schema_result.to_pydict()["schema"]
 
-def _get_filepaths(path: str):
-    fs = get_filesystem_from_path(path)
-    protocol = get_protocol_from_path(path)
-    if fs.isdir(path):
-        return [f"{protocol}://{path}" if protocol != "file" else path for path in fs.ls(path)]
-    elif fs.isfile(path):
-        return [path]
-    return [f"{protocol}://{path}" if protocol != "file" else path for path in fs.expand_path(path, recursive=True)]
+    # TODO: infer schema from all sampled schemas instead of just taking the first one
+    schema = sampled_schemas[0]
+    schema = schema.resolve()
+
+    # Return a TabularFilesScan node that will scan from the globbed filepaths filepaths
+    return logical_plan.TabularFilesScan(
+        schema=schema,
+        predicate=None,
+        columns=None,
+        source_info=source_info,
+        filepaths_child=filepath_plan,
+        filepaths_column_name=partition_set_factory.FILEPATH_COLUMN_NAME,
+    )
 
 
 class DataFrame:
@@ -83,6 +96,7 @@ class DataFrame:
         """
         self.__plan = plan
         self._result_cache: PartitionCacheEntry | None = None
+        self._preview = DataFramePreview(preview_partition=None, dataframe_num_rows=None)
 
     @property
     def _plan(self) -> logical_plan.LogicalPlan:
@@ -160,7 +174,7 @@ class DataFrame:
         """Executes and displays the executed dataframe as a table
 
         Args:
-            n: number of rows to show. Defaults to -1.
+            n: number of rows to show. Defaults to None which indicates showing the entire Dataframe.
 
         Returns:
             DataFrameDisplay: object that has a rich tabular display
@@ -172,14 +186,27 @@ class DataFrame:
         if n is not None:
             df = df.limit(n)
 
-        execution_result = df.to_pandas()
-        return DataFrameDisplay(execution_result, df.schema())
+        df.collect(num_preview_rows=n)
+        result = df._result
+        assert result is not None
+
+        # If showing all rows, then we can use the resulting DataFramePreview's dataframe_num_rows since no limit was applied
+        dataframe_num_rows = df._preview.dataframe_num_rows if n is None else None
+
+        preview = DataFramePreview(
+            preview_partition=df._preview.preview_partition,
+            dataframe_num_rows=dataframe_num_rows,
+        )
+
+        return DataFrameDisplay(preview, df.schema())
 
     def __repr__(self) -> str:
-        return self.schema().__repr__()
+        display = DataFrameDisplay(self._preview, self.schema())
+        return display.__repr__()
 
     def _repr_html_(self) -> str:
-        return self.schema()._repr_html_()
+        display = DataFrameDisplay(self._preview, self.schema())
+        return display._repr_html_()
 
     ###
     # Creation methods
@@ -260,17 +287,22 @@ class DataFrame:
         return cls(plan)
 
     @classmethod
-    def from_json(
+    def from_json(cls, *args, **kwargs) -> DataFrame:
+        warnings.warn(f"DataFrame.from_json will be deprecated in 0.1.0 in favor of DataFrame.read_json")
+        return cls.read_json(*args, **kwargs)
+
+    @classmethod
+    def read_json(
         cls,
         path: str,
     ) -> DataFrame:
         """Creates a DataFrame from line-delimited JSON file(s)
 
         Example:
-            >>> df = DataFrame.from_json("/path/to/file.json")
-            >>> df = DataFrame.from_json("/path/to/directory")
-            >>> df = DataFrame.from_json("/path/to/files-*.json")
-            >>> df = DataFrame.from_json("s3://path/to/files-*.json")
+            >>> df = DataFrame.read_json("/path/to/file.json")
+            >>> df = DataFrame.read_json("/path/to/directory")
+            >>> df = DataFrame.read_json("/path/to/files-*.json")
+            >>> df = DataFrame.read_json("s3://path/to/files-*.json")
 
         Args:
             path (str): Path to JSON files (allows for wildcards)
@@ -278,23 +310,35 @@ class DataFrame:
         returns:
             DataFrame: parsed DataFrame
         """
-        filepaths = _get_filepaths(path)
 
-        if len(filepaths) == 0:
-            raise ValueError(f"No JSON files found at {path}")
+        def get_schema(filepath: str) -> ExpressionList:
+            return vPartition.from_json(
+                filepath,
+                partition_id=0,
+                schema_options=vPartitionSchemaInferenceOptions(
+                    schema=None,
+                    inference_column_names=None,  # has no effect on inferring schema from JSON
+                ),
+                read_options=vPartitionReadOptions(
+                    num_rows=100,  # sample 100 rows for inferring schema
+                    column_names=None,  # read all columns
+                ),
+            ).get_unresolved_col_expressions()
 
-        schema = _sample_with_pyarrow(json.read_json, filepaths[0])
-
-        plan = logical_plan.Scan(
-            schema=schema,
-            predicate=None,
-            columns=None,
-            source_info=JSONSourceInfo(filepaths=filepaths),
+        plan = _get_tabular_files_scan(
+            path,
+            get_schema,
+            JSONSourceInfo(),
         )
         return cls(plan)
 
     @classmethod
-    def from_csv(
+    def from_csv(cls, *args, **kwargs) -> DataFrame:
+        warnings.warn(f"DataFrame.from_csv will be deprecated in 0.1.0 in favor of DataFrame.read_csv")
+        return cls.read_csv(*args, **kwargs)
+
+    @classmethod
+    def read_csv(
         cls,
         path: str,
         has_headers: bool = True,
@@ -304,10 +348,10 @@ class DataFrame:
         """Creates a DataFrame from CSV file(s)
 
         Example:
-            >>> df = DataFrame.from_csv("/path/to/file.csv")
-            >>> df = DataFrame.from_csv("/path/to/directory")
-            >>> df = DataFrame.from_csv("/path/to/files-*.csv")
-            >>> df = DataFrame.from_csv("s3://path/to/files-*.csv")
+            >>> df = DataFrame.read_csv("/path/to/file.csv")
+            >>> df = DataFrame.read_csv("/path/to/directory")
+            >>> df = DataFrame.read_csv("/path/to/files-*.csv")
+            >>> df = DataFrame.read_csv("s3://path/to/files-*.csv")
 
         Args:
             path (str): Path to CSV (allows for wildcards)
@@ -318,33 +362,31 @@ class DataFrame:
         returns:
             DataFrame: parsed DataFrame
         """
-        filepaths = _get_filepaths(path)
 
-        if len(filepaths) == 0:
-            raise ValueError(f"No CSV files found at {path}")
-
-        schema = _sample_with_pyarrow(
-            partial(
-                csv.read_csv,
-                parse_options=csv.ParseOptions(
+        def get_schema(filepath: str) -> ExpressionList:
+            return vPartition.from_csv(
+                path=filepath,
+                partition_id=0,
+                csv_options=vPartitionParseCSVOptions(
                     delimiter=delimiter,
+                    has_headers=has_headers,
+                    skip_rows_before_header=0,
+                    skip_rows_after_header=0,
                 ),
-                read_options=csv.ReadOptions(
-                    # Column names will be read from the first CSV row if column_names is None/empty and has_headers
-                    autogenerate_column_names=(not has_headers) and (column_names is None),
-                    column_names=column_names,
-                    # If user specifies that CSV has headers, and also provides column names, we skip the header row
-                    skip_rows_after_names=1 if has_headers and column_names is not None else 0,
+                schema_options=vPartitionSchemaInferenceOptions(
+                    schema=None,
+                    inference_column_names=column_names,  # pass in user-provided column names
                 ),
-            ),
-            filepaths[0],
-        )
-        plan = logical_plan.Scan(
-            schema=schema,
-            predicate=None,
-            columns=None,
-            source_info=CSVSourceInfo(
-                filepaths=filepaths,
+                read_options=vPartitionReadOptions(
+                    num_rows=100,  # sample 100 rows for schema inference
+                    column_names=None,  # read all columns
+                ),
+            ).get_unresolved_col_expressions()
+
+        plan = _get_tabular_files_scan(
+            path,
+            get_schema,
+            CSVSourceInfo(
                 delimiter=delimiter,
                 has_headers=has_headers,
             ),
@@ -352,14 +394,19 @@ class DataFrame:
         return cls(plan)
 
     @classmethod
-    def from_parquet(cls, path: str) -> DataFrame:
+    def from_parquet(cls, *args, **kwargs) -> DataFrame:
+        warnings.warn(f"DataFrame.from_parquet will be deprecated in 0.1.0 in favor of DataFrame.read_parquet")
+        return cls.read_parquet(*args, **kwargs)
+
+    @classmethod
+    def read_parquet(cls, path: str) -> DataFrame:
         """Creates a DataFrame from Parquet file(s)
 
         Example:
-            >>> df = DataFrame.from_parquet("/path/to/file.parquet")
-            >>> df = DataFrame.from_parquet("/path/to/directory")
-            >>> df = DataFrame.from_parquet("/path/to/files-*.parquet")
-            >>> df = DataFrame.from_parquet("s3://path/to/files-*.parquet")
+            >>> df = DataFrame.read_parquet("/path/to/file.parquet")
+            >>> df = DataFrame.read_parquet("/path/to/directory")
+            >>> df = DataFrame.read_parquet("/path/to/files-*.parquet")
+            >>> df = DataFrame.read_parquet("s3://path/to/files-*.parquet")
 
         Args:
             path (str): Path to Parquet file (allows for wildcards)
@@ -367,29 +414,25 @@ class DataFrame:
         returns:
             DataFrame: parsed DataFrame
         """
-        filepaths = _get_filepaths(path)
 
-        if len(filepaths) == 0:
-            raise ValueError(f"No Parquet files found at {path}")
+        def get_schema(filepath: str) -> ExpressionList:
+            return vPartition.from_parquet(
+                filepath,
+                partition_id=0,
+                schema_options=vPartitionSchemaInferenceOptions(
+                    schema=None,
+                    inference_column_names=None,  # has no effect on schema inferencing Parquet
+                ),
+                read_options=vPartitionReadOptions(
+                    num_rows=0,  # sample 0 rows since Parquet has metadata
+                    column_names=None,  # read all columns
+                ),
+            ).get_unresolved_col_expressions()
 
-        filepath = filepaths[0]
-        fs = get_filesystem_from_path(filepath)
-        with fs.open(filepath, "rb") as f:
-            # Read first Parquet file to ascertain schema
-            schema = ExpressionList(
-                [
-                    ColumnExpression(field.name, expr_type=ExpressionType.from_arrow_type(field.type))
-                    for field in papq.ParquetFile(f).metadata.schema.to_arrow_schema()
-                ]
-            )
-
-        plan = logical_plan.Scan(
-            schema=schema,
-            predicate=None,
-            columns=None,
-            source_info=ParquetSourceInfo(
-                filepaths=filepaths,
-            ),
+        plan = _get_tabular_files_scan(
+            path,
+            get_schema,
+            ParquetSourceInfo(),
         )
         return cls(plan)
 
@@ -812,6 +855,8 @@ class DataFrame:
             "sum": Expression._sum,
             "count": Expression._count,
             "mean": Expression._mean,
+            "list": Expression._list,
+            "concat": Expression._concat,
             "min": Expression._min,
             "max": Expression._max,
         }
@@ -827,13 +872,21 @@ class DataFrame:
 
         intermediate_ops = {
             "sum": ("sum",),
+            "list": ("list",),
             "count": ("count",),
             "mean": ("sum", "count"),
             "min": ("min",),
             "max": ("max",),
         }
 
-        reduction_ops = {"sum": ("sum",), "count": ("sum",), "mean": ("sum", "sum"), "min": ("min",), "max": ("max",)}
+        reduction_ops = {
+            "sum": ("sum",),
+            "list": ("concat",),
+            "count": ("sum",),
+            "mean": ("sum", "sum"),
+            "min": ("min",),
+            "max": ("max",),
+        }
 
         finalizer_ops_funcs = {"mean": lambda x, y: (x + 0.0) / (y + 0.0)}
 
@@ -1001,18 +1054,47 @@ class DataFrame:
         """
         return GroupedDataFrame(self, self.__column_input_to_expression(group_by))
 
-    def collect(self) -> DataFrame:
-        """Computes LogicalPlan to materialize DataFrame. This is a blocking operation.
-
-        Returns:
-            DataFrame: DataFrame with cached results.
-        """
+    def _materialize_results(self) -> None:
+        """Materializes the results of for this DataFrame and hold a pointer to the results."""
         context = get_context()
         if self._result is None:
             self._result_cache = context.runner().run(self._plan)
             result = self._result
             assert result is not None
             result.wait()
+
+    def collect(self, num_preview_rows: int | None = 10) -> DataFrame:
+        """Computes LogicalPlan to materialize DataFrame. This is a blocking operation.
+
+        Args:
+            num_preview_rows: Number of rows to preview. Defaults to 10
+
+        Returns:
+            DataFrame: DataFrame with materialized results.
+        """
+        self._materialize_results()
+
+        assert self._result is not None
+        dataframe_len = len(self._result)
+        requested_rows = dataframe_len if num_preview_rows is None else num_preview_rows
+
+        # Build a DataFramePreview and cache it if we need to
+        if self._preview.preview_partition is None or len(self._preview.preview_partition) < requested_rows:
+
+            # Add a limit onto self and materialize limited data
+            preview_df = self
+            if num_preview_rows is not None:
+                preview_df = preview_df.limit(num_preview_rows)
+            preview_df._materialize_results()
+            preview_results = preview_df._result
+            assert preview_results is not None
+
+            preview_partition = preview_results._get_merged_vpartition()
+            self._preview = DataFramePreview(
+                preview_partition=preview_partition,
+                dataframe_num_rows=dataframe_len,
+            )
+
         return self
 
     def to_pandas(self) -> pandas.DataFrame:
@@ -1056,6 +1138,8 @@ class DataFrame:
         Returns:
             RayDataset: Ray dataset
         """
+        from daft.runners.ray_runner import RayPartitionSet
+
         self.collect()
         partition_set = self._result
         assert partition_set is not None
@@ -1074,6 +1158,9 @@ class GroupedDataFrame:
             if e.resolved_type() == ExpressionType.null():
                 raise ExpressionTypeError(f"Cannot groupby on null type expression: {e}")
 
+    def __getitem__(self, item: slice | int | str | Iterable[str | int]) -> ColumnExpression | DataFrame:
+        return self.df.__getitem__(item)
+
     def sum(self, *cols: ColumnInputType) -> DataFrame:
         """Perform grouped sum on this GroupedDataFrame.
 
