@@ -19,7 +19,13 @@ from typing import (
 
 import numpy as np
 import pandas as pd
-import polars as pl
+
+_POLARS_AVAILABLE = True
+try:
+    import polars as pl
+except ImportError:
+    _POLARS_AVAILABLE = False
+
 import pyarrow as pa
 
 from daft.errors import ExpressionTypeError
@@ -101,8 +107,35 @@ class ExpressionExecutor:
         elif isinstance(expr, UdfExpression):
             eval_args = tuple(self.eval(a, operands) for a in expr._args)
             eval_kwargs = {kw: self.eval(a, operands) for kw, a in expr._kwargs.items()}
-            result = expr.eval_blocks(*eval_args, **eval_kwargs)
-            return result
+
+            results = expr._func(*eval_args, **eval_kwargs)
+
+            if ExpressionType.is_py(expr._func_ret_type):
+                return PyListDataBlock(results if isinstance(results, list) else list(results))
+
+            # Convert these user-provided types to pa.ChunkedArray for making blocks
+            if isinstance(results, pa.Array):
+                results = pa.chunked_array([results])
+            elif _POLARS_AVAILABLE and isinstance(results, pl.Series):
+                results = pa.chunked_array([results.to_arrow()])
+            elif isinstance(results, np.ndarray):
+                results = pa.chunked_array([pa.array(results)])
+            elif isinstance(results, pd.Series):
+                results = pa.chunked_array([pa.array(results)])
+            elif isinstance(results, list):
+                results = pa.chunked_array([pa.array(results)])
+
+            if not isinstance(results, pa.ChunkedArray):
+                raise NotImplementedError(f"Cannot make block for data of type: {type(results)}")
+
+            # Explcitly cast results of the UDF here for these primitive types:
+            #   1. Ensures that all blocks across all partitions will have the same underlying Arrow type after the UDF
+            #   2. Ensures that any null blocks from empty partitions or partitions with all nulls have the correct type
+            assert isinstance(expr._func_ret_type, PrimitiveExpressionType)
+            expected_arrow_type = expr._func_ret_type.to_arrow_type()
+            results = results.cast(expected_arrow_type)
+
+            return ArrowDataBlock(results)
         elif isinstance(expr, AsPyExpression):
             raise NotImplementedError("AsPyExpressions need to be evaluated with a method call")
         else:
@@ -240,6 +273,8 @@ class Expression(TreeNode["Expression"]):
     _sum = partialmethod(_unary_op, OperatorEnum.SUM)
     _count = partialmethod(_unary_op, OperatorEnum.COUNT)
     _mean = partialmethod(_unary_op, OperatorEnum.MEAN)
+    _list = partialmethod(_unary_op, OperatorEnum.LIST)
+    _concat = partialmethod(_unary_op, OperatorEnum.CONCAT)
     _min = partialmethod(_unary_op, OperatorEnum.MIN)
     _max = partialmethod(_unary_op, OperatorEnum.MAX)
     _explode = partialmethod(_unary_op, OperatorEnum.EXPLODE)
@@ -624,36 +659,6 @@ class UdfExpression(Expression):
             return f"{func_name}({args}, {kwargs})"
         return f"{func_name}({args})"
 
-    def eval_blocks(self, *args: DataBlock, **kwargs: DataBlock) -> DataBlock:
-        results = self._func(*args, **kwargs)
-
-        if ExpressionType.is_py(self._func_ret_type):
-            return PyListDataBlock(results if isinstance(results, list) else list(results))
-
-        # Convert these user-provided types to pa.ChunkedArray for making blocks
-        if isinstance(results, pa.Array):
-            results = pa.chunked_array([results])
-        elif isinstance(results, pl.Series):
-            results = pa.chunked_array([results.to_arrow()])
-        elif isinstance(results, np.ndarray):
-            results = pa.chunked_array([pa.array(results)])
-        elif isinstance(results, pd.Series):
-            results = pa.chunked_array([pa.array(results)])
-        elif isinstance(results, list):
-            results = pa.chunked_array([pa.array(results)])
-
-        if not isinstance(results, pa.ChunkedArray):
-            raise NotImplementedError(f"Cannot make block for data of type: {type(results)}")
-
-        # Explcitly cast results of the UDF here for these primitive types:
-        #   1. Ensures that all blocks across all partitions will have the same underlying Arrow type after the UDF
-        #   2. Ensures that any null blocks from empty partitions or partitions with all nulls have the correct type
-        assert isinstance(self._func_ret_type, PrimitiveExpressionType)
-        expected_arrow_type = self._func_ret_type.to_arrow_type()
-        results = results.cast(expected_arrow_type)
-
-        return ArrowDataBlock(results)
-
     def _is_eq_local(self, other: Expression) -> bool:
         return (
             isinstance(other, UdfExpression)
