@@ -102,7 +102,7 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
         def helper(node: LogicalPlan, depth: int = 0, index: int = 0, prefix: str = "", header: str = ""):
             children: list[LogicalPlan] = node._children()
             obj_repr_lines = repr(node).splitlines()
-            builder.append(f"{prefix}{header}{obj_repr_lines[0]}\n")
+            builder.append(f"{header}{obj_repr_lines[0]}\n")
 
             if len(children) > 0:
                 body_prefix = prefix + "│"
@@ -118,9 +118,9 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
                     has_grandchild = len(child._children()) > 0
 
                     if has_grandchild:
-                        header = "├──"
+                        header = prefix + "├──"
                     else:
-                        header = "└──"
+                        header = prefix + "└──"
 
                     helper(child, depth=depth, index=index + 1, prefix=prefix, header=header)
             else:
@@ -136,11 +136,16 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
                         final_header = "───"
 
                     position = len(children) - i
-                    header = ("\b\b\b" * position) + connector + (middle_child_header * (position - 1)) + final_header
                     if i != len(children) - 1:
                         next_child_prefix = prefix + ("   │  " * (position - 1))
                     else:
                         next_child_prefix = prefix + "      "
+                    header = (
+                        next_child_prefix[: -3 * position]
+                        + connector
+                        + (middle_child_header * (position - 1))
+                        + final_header
+                    )
 
                     helper(child, depth=depth + 1, index=i, prefix=next_child_prefix, header=header)
 
@@ -179,7 +184,7 @@ class BinaryNode(LogicalPlan):
     ...
 
 
-class Scan(LogicalPlan):
+class TabularFilesScan(UnaryNode):
     def __init__(
         self,
         *,
@@ -187,9 +192,11 @@ class Scan(LogicalPlan):
         source_info: SourceInfo,
         predicate: ExpressionList | None = None,
         columns: list[str] | None = None,
+        filepaths_child: LogicalPlan,
+        filepaths_column_name: str,
     ) -> None:
         schema = schema.resolve()
-        pspec = PartitionSpec(scheme=PartitionScheme.UNKNOWN, num_partitions=source_info.get_num_partitions())
+        pspec = PartitionSpec(scheme=PartitionScheme.UNKNOWN, num_partitions=filepaths_child.num_partitions())
         super().__init__(schema, partition_spec=pspec, op_level=OpLevel.PARTITION)
 
         if predicate is not None:
@@ -206,6 +213,18 @@ class Scan(LogicalPlan):
         self._columns = self._schema
         self._source_info = source_info
 
+        # TabularFilsScan has a single child node that provides the filepaths to read from.
+        assert (
+            filepaths_child.schema().get_expression_by_name(filepaths_column_name) is not None
+        ), f"TabularFileScan requires a child with '{filepaths_column_name}' column"
+        self._register_child(filepaths_child)
+        self._filepaths_column_name = filepaths_column_name
+
+    @property
+    def _filepaths_child(self) -> LogicalPlan:
+        child = self._children()[0]
+        return child
+
     def schema(self) -> ExpressionList:
         return self._output_schema
 
@@ -216,28 +235,40 @@ class Scan(LogicalPlan):
         return ResourceRequest.default()
 
     def required_columns(self) -> ExpressionList:
-        return self._predicate.required_columns()
+        filepaths_col = self._filepaths_child.schema().get_expression_by_name(self._filepaths_column_name)
+        return ExpressionList.union(self._predicate.required_columns(), ExpressionList([filepaths_col]))
 
     def _local_eq(self, other: Any) -> bool:
         return (
-            isinstance(other, Scan)
+            isinstance(other, TabularFilesScan)
             and self.schema() == other.schema()
             and self._predicate == other._predicate
             and self._columns == other._columns
             and self._source_info == other._source_info
+            and self._filepaths_column_name == other._filepaths_column_name
         )
 
     def rebuild(self) -> LogicalPlan:
-        return Scan(
+        child = self._filepaths_child.rebuild()
+        return TabularFilesScan(
             schema=self.schema().unresolve(),
             source_info=self._source_info,
             predicate=self._predicate.unresolve() if self._predicate is not None else None,
             columns=self._column_names,
+            filepaths_child=child,
+            filepaths_column_name=self._filepaths_column_name,
         )
 
     def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
-        assert len(new_children) == 0
-        return self
+        assert len(new_children) == 1
+        return TabularFilesScan(
+            schema=self.schema(),
+            source_info=self._source_info,
+            predicate=self._predicate,
+            columns=self._column_names,
+            filepaths_child=new_children[0],
+            filepaths_column_name=self._filepaths_column_name,
+        )
 
 
 class InMemoryScan(UnaryNode):
