@@ -9,6 +9,7 @@ import random
 from abc import abstractmethod
 from functools import partial
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     ClassVar,
@@ -22,8 +23,13 @@ from typing import (
     cast,
 )
 
+_POLARS_AVAILABLE = True
+try:
+    import polars as pl
+except ImportError:
+    _POLARS_AVAILABLE = False
+
 import numpy as np
-import polars as pl
 import pyarrow as pa
 import pyarrow.compute as pac
 from pandas.core.reshape.merge import get_join_indexers
@@ -33,6 +39,9 @@ from daft.internal.kernels.hashing import hash_chunked_array
 from daft.internal.kernels.search_sorted import search_sorted
 from daft.types import ExpressionType, PrimitiveExpressionType, PythonExpressionType
 
+if TYPE_CHECKING:
+    import polars as pl
+
 # A type representing some Python scalar (non-series/array like object)
 PyScalar = Any
 
@@ -40,8 +49,6 @@ ArrType = TypeVar("ArrType", bound=collections.abc.Sequence)
 UnaryFuncType = Callable[[ArrType], ArrType]
 BinaryFuncType = Callable[[ArrType, ArrType], ArrType]
 
-NUMPY_MINOR_VERSION = int(np.version.version.split(".")[1])
-
 
 def zip_blocks_as_py(*blocks: DataBlock) -> Iterator[tuple[Any, ...]]:
     """Utility to zip the data of blocks together, returning a row-based iterator. This utility
@@ -109,8 +116,13 @@ class DataBlock(Generic[ArrType]):
     def to_numpy(self) -> np.ndarray:
         raise NotImplementedError()
 
-    @abstractmethod
     def to_polars(self) -> PyScalar | pl.Series:
+        if not _POLARS_AVAILABLE:
+            raise ImportError("Polars library is not available")
+        return self._to_polars()
+
+    @abstractmethod
+    def _to_polars(self) -> PyScalar | pl.Series:
         raise NotImplementedError()
 
     @abstractmethod
@@ -262,7 +274,7 @@ class DataBlock(Generic[ArrType]):
     @classmethod
     def argsort(cls, blocks: list[DataBlock], descending: list[bool] | None = None) -> DataBlock[ArrowArrType]:
         first_type = type(blocks[0])
-        assert all(type(b) == first_type for b in blocks), "all block types must match"
+        assert all(type(b) == first_type for b in blocks), f"all block types must match: {[type(b) for b in blocks]}"
         size = len(blocks)
         if descending is None:
             descending = [False for _ in range(size)]
@@ -282,8 +294,10 @@ class DataBlock(Generic[ArrType]):
         cls, sorted_blocks: list[DataBlock], keys: list[DataBlock], input_reversed: list[bool] | None = None
     ) -> DataBlock[ArrowArrType]:
         first_type = type(sorted_blocks[0])
-        assert all(type(b) == first_type for b in sorted_blocks), "all block types must match"
-        assert all(type(b) == first_type for b in keys), "all block types must match"
+        assert all(
+            type(b) == first_type for b in sorted_blocks
+        ), f"all block types must match: {[type(b) for b in sorted_blocks]}"
+        assert all(type(b) == first_type for b in keys), f"all block types must match: {[type(b) for b in keys]}"
 
         size = len(sorted_blocks)
         if input_reversed is None:
@@ -302,7 +316,7 @@ class DataBlock(Generic[ArrType]):
     def merge_blocks(cls, blocks: list[DataBlock[ArrType]]) -> DataBlock[ArrType]:
         assert len(blocks) > 0, "no blocks"
         first_type = type(blocks[0])
-        assert all(type(b) == first_type for b in blocks), "all block types must match"
+        assert all(type(b) == first_type for b in blocks), f"all block types must match: {[type(b) for b in blocks]}"
         return first_type._merge_blocks(blocks)
 
     @abstractmethod
@@ -326,15 +340,9 @@ class DataBlock(Generic[ArrType]):
     ) -> tuple[list[DataBlock[ArrType]], list[DataBlock[ArrType]]]:
         assert len(group_by) > 0, "no blocks"
         assert len(to_agg) == len(agg_ops)
-        if len(to_agg) > 0:
-            first_type = type(to_agg[0])
-            assert all(type(b) == first_type for b in to_agg), "all block types must match"
-        else:
-            first_type = type(group_by[0])
+        assert all(type(b) == ArrowDataBlock for b in group_by), "Groupby columns must all be Arrow"
 
-        assert all(type(b) == first_type for b in group_by), "all block types must match"
-
-        return first_type._group_by_agg(group_by, to_agg, agg_ops)
+        return ArrowDataBlock._group_by_agg(group_by, to_agg, agg_ops)
 
     def identity(self) -> DataBlock[ArrType]:
         return self
@@ -412,7 +420,7 @@ class PyListDataBlock(DataBlock[List[T]]):
             res[i] = obj
         return res
 
-    def to_polars(self) -> PyScalar | pl.Series:
+    def _to_polars(self) -> PyScalar | pl.Series:
         if self.is_scalar():
             return self.data
         return pl.Series(self.data, dtype=pl.datatypes.Object)
@@ -458,6 +466,10 @@ class PyListDataBlock(DataBlock[List[T]]):
         return DataBlock.make_block(np.array(hashes))
 
     def agg(self, op: str) -> DataBlock[ArrowArrType]:
+        if op == "concat":
+            if len(self) == 0:
+                return PyListDataBlock([])
+            return PyListDataBlock([sum(self.data, [])])
         raise NotImplementedError("Aggregations on Python objects is not implemented yet")
 
     @staticmethod
@@ -514,7 +526,7 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
             return self.data.as_py()
         return self.data.to_numpy()
 
-    def to_polars(self) -> PyScalar | pl.Series:
+    def _to_polars(self) -> PyScalar | pl.Series:
         if isinstance(self.data, pa.Scalar):
             return self.data.as_py()
         return pl.Series(self.data)
@@ -603,12 +615,23 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
 
     @staticmethod
     def _merge_blocks(blocks: list[DataBlock[ArrowArrType]]) -> DataBlock[ArrowArrType]:
-        block_types = {block.data.type for block in blocks}
-        assert len(block_types) == 1, f"Unable to merge blocks of different types: {block_types}"
+        unique_block_types = {block.data.type for block in blocks}
+        blocks_to_merge = blocks
+        if len(unique_block_types) == 1:
+            default_type = unique_block_types.pop()
+        elif len(unique_block_types) == 2:
+            if any(pa.types.is_null(type) for type in unique_block_types):
+                default_type = [type for type in unique_block_types if not pa.types.is_null(type)][0]
+                blocks_to_merge = [ArrowDataBlock(block.data.cast(default_type)) for block in blocks]
+            else:
+                raise ValueError(f"can not merge different block types {unique_block_types}")
+        else:
+            raise ValueError(f"can not merge different block types {unique_block_types}")
+
         all_chunks = []
-        for block in blocks:
+        for block in blocks_to_merge:
             all_chunks.extend(block.data.chunks)
-        return DataBlock.make_block(pa.chunked_array(all_chunks, type=block_types.pop()))
+        return DataBlock.make_block(pa.chunked_array(all_chunks, type=default_type))
 
     def _make_empty(self) -> DataBlock[ArrowArrType]:
         return DataBlock.make_block(data=pa.chunked_array([[]], type=self.data.type))
@@ -635,6 +658,10 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
             if len(self) == 0:
                 return ArrowDataBlock(data=pa.chunked_array([[]], type=pa.float64()))
             return ArrowDataBlock(data=pa.chunked_array([[pac.mean(self.data).as_py()]], type=pa.float64()))
+        elif op == "list":
+            if len(self) == 0:
+                return PyListDataBlock([])
+            return PyListDataBlock([self.data.to_pylist()])
         elif op == "count":
             if len(self) == 0:
                 return ArrowDataBlock(data=pa.chunked_array([[]], type=pa.int64()))
@@ -659,6 +686,7 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
         arrs = group_arrs + agg_arrs
         group_names = [f"g_{i}" for i in range(len(group_by))]
         agg_names = [f"a_{i}" for i in range(len(to_agg))]
+
         table = pa.table(arrs, names=group_names + agg_names)
         pl_table = pl.from_arrow(table, rechunk=True)
 
@@ -672,6 +700,27 @@ class ArrowDataBlock(DataBlock[ArrowArrType]):
             elif op == "mean":
                 exprs.append(pl.mean(an))
                 agg_expected_arrow_type.append(pa.float64())
+            elif op == "list":
+                exprs.append(pl.list(an))
+                agg_expected_arrow_type.append(pa.list_(arr.type))
+            elif op == "concat":
+                if len(arr) == 0:
+                    # If the column is empty, explode() will not work due to type information being missing.
+                    # Manually construct the result in that case (by passsing through the empty column).
+                    exprs.append(pl.col(an))
+                    agg_expected_arrow_type.append(pa.list_(table[an].type))
+                elif table[an].type == pa.list_(pa.null()):
+                    # Force a polars cast to list[i8]
+                    # (since polars cannot aggregate list[null])
+                    # TODO: File polars issue.
+                    exprs.append(pl.col(an).cast(pl.List(pl.Int8())).explode().list())
+                    # Polars convers Polars list[null] to Arrow list[i8].
+                    # TODO: File polars issue.
+                    agg_expected_arrow_type.append(pa.list_(pa.int8()))
+                else:
+                    # Regular case.
+                    exprs.append(pl.col(an).explode().list())
+                    agg_expected_arrow_type.append(table[an].type)
             elif op == "min":
                 exprs.append(pl.min(an))
                 agg_expected_arrow_type.append(arr.type)
@@ -874,6 +923,8 @@ class ArrowEvaluator(OperatorEvaluator["ArrowDataBlock"]):
     # They exist on the Evaluator only to provide correct typing information
     SUM = ArrowDataBlock.identity
     MEAN = ArrowDataBlock.identity
+    LIST = ArrowDataBlock.identity
+    CONCAT = ArrowDataBlock.identity
     MIN = ArrowDataBlock.identity
     MAX = ArrowDataBlock.identity
     COUNT = ArrowDataBlock.identity
@@ -987,6 +1038,8 @@ class PyListEvaluator(OperatorEvaluator["PyListDataBlock"]):
     # They exist on the Evaluator only to provide correct typing information
     SUM = PyListDataBlock.identity
     MEAN = PyListDataBlock.identity
+    LIST = PyListDataBlock.identity
+    CONCAT = PyListDataBlock.identity
     MIN = PyListDataBlock.identity
     MAX = PyListDataBlock.identity
     COUNT = PyListDataBlock.identity
