@@ -4,16 +4,21 @@ import multiprocessing
 from dataclasses import dataclass
 from typing import Callable, ClassVar
 
+import psutil
+
 from daft.execution.execution_plan import ExecutionPlan
 from daft.execution.logical_op_runners import (
     LogicalGlobalOpRunner,
     LogicalPartitionOpRunner,
     ReduceType,
 )
+from daft.expressions import ColumnExpression
+from daft.filesystem import glob_path
 from daft.internal.gpu import cuda_device_count
 from daft.internal.rule_runner import FixedPointPolicy, Once, RuleBatch, RuleRunner
-from daft.logical.logical_plan import LogicalPlan
+from daft.logical import logical_plan
 from daft.logical.optimizer import (
+    DropProjections,
     DropRepartition,
     FoldProjections,
     PruneColumns,
@@ -21,11 +26,13 @@ from daft.logical.optimizer import (
     PushDownLimit,
     PushDownPredicates,
 )
+from daft.logical.schema import ExpressionList
 from daft.resource_request import ResourceRequest
 from daft.runners.partitioning import (
     PartID,
     PartitionCacheEntry,
     PartitionSet,
+    PartitionSetFactory,
     vPartition,
 )
 from daft.runners.profiler import profiler
@@ -38,17 +45,21 @@ from daft.runners.shuffle_ops import (
     Shuffler,
     SortOp,
 )
+from daft.types import ExpressionType
 
 
 @dataclass
 class LocalPartitionSet(PartitionSet[vPartition]):
     _partitions: dict[PartID, vPartition]
 
-    def _get_all_vpartitions(self) -> list[vPartition]:
-        partition_ids = sorted(list(self._partitions.keys()))
-        assert partition_ids[0] == 0
-        assert partition_ids[-1] + 1 == len(partition_ids)
-        return [self._partitions[pid] for pid in partition_ids]
+    def items(self) -> list[tuple[PartID, vPartition]]:
+        return sorted(self._partitions.items())
+
+    def _get_merged_vpartition(self) -> vPartition:
+        ids_and_partitions = self.items()
+        assert ids_and_partitions[0][0] == 0
+        assert ids_and_partitions[-1][0] + 1 == len(ids_and_partitions)
+        return vPartition.merge_partitions([part for id, part in ids_and_partitions], verify_partition_id=False)
 
     def get_partition(self, idx: PartID) -> vPartition:
         return self._partitions[idx]
@@ -76,6 +87,24 @@ class LocalPartitionSet(PartitionSet[vPartition]):
         pass
 
 
+class LocalPartitionSetFactory(PartitionSetFactory[vPartition]):
+    def glob_filepaths(
+        self,
+        source_path: str,
+    ) -> tuple[LocalPartitionSet, ExpressionList]:
+        filepaths = glob_path(source_path)
+        schema = ExpressionList([ColumnExpression(self.FILEPATH_COLUMN_NAME, ExpressionType.string())]).resolve()
+        pset = LocalPartitionSet(
+            {
+                i: vPartition.from_pydict(
+                    data={self.FILEPATH_COLUMN_NAME: [filepaths[i]]}, schema=schema, partition_id=i
+                )
+                for i in range(len(filepaths))  # Hardcoded to 1 path per partition
+            },
+        )
+        return pset, schema
+
+
 class PyRunnerSimpleShuffler(Shuffler):
     def run(self, input: PartitionSet, num_target_partitions: int) -> PartitionSet:
         map_args = self._map_args if self._map_args is not None else {}
@@ -116,7 +145,7 @@ class LocalLogicalPartitionOpRunner(LogicalPartitionOpRunner):
     def run_node_list(
         self,
         inputs: dict[int, PartitionSet],
-        nodes: list[LogicalPlan],
+        nodes: list[logical_plan.LogicalPlan],
         num_partitions: int,
         resource_request: ResourceRequest,
     ) -> PartitionSet:
@@ -169,15 +198,18 @@ class PyRunner(Runner):
                 RuleBatch(
                     "PushDownLimitsAndRepartitions",
                     FixedPointPolicy(3),
-                    [PushDownLimit(), DropRepartition()],
+                    [PushDownLimit(), DropRepartition(), DropProjections()],
                 ),
             ]
         )
 
-    def optimize(self, plan: LogicalPlan) -> LogicalPlan:
+    def optimize(self, plan: logical_plan.LogicalPlan) -> logical_plan.LogicalPlan:
         return self._optimizer.optimize(plan)
 
-    def run(self, plan: LogicalPlan) -> PartitionCacheEntry:
+    def partition_set_factory(self) -> PartitionSetFactory:
+        return LocalPartitionSetFactory()
+
+    def run(self, plan: logical_plan.LogicalPlan) -> PartitionCacheEntry:
         optimized_plan = self.optimize(plan)
 
         exec_plan = ExecutionPlan.plan_from_logical(optimized_plan)
@@ -194,6 +226,13 @@ class PyRunner(Runner):
                 raise RuntimeError(
                     f"Requested {resource_request.num_gpus} GPUs but found only {cuda_device_count()} available"
                 )
+            if (
+                resource_request.memory_bytes is not None
+                and resource_request.memory_bytes > psutil.virtual_memory().total
+            ):
+                raise RuntimeError(
+                    f"Requested {resource_request.memory_bytes} bytes of memory but found only {psutil.virtual_memory().total} available"
+                )
         partition_intermediate_results: dict[int, PartitionSet] = {}
         with profiler("profile.json"):
             for exec_op in exec_plan.execution_ops:
@@ -202,7 +241,6 @@ class PyRunner(Runner):
                 input_partition_set = {nid: partition_intermediate_results[nid] for nid in data_deps}
 
                 if exec_op.is_global_op:
-                    input_partition_set = {nid: partition_intermediate_results[nid] for nid in data_deps}
                     result_partition_set = self._global_op_runner.run_node_list(
                         input_partition_set, exec_op.logical_ops
                     )
@@ -218,5 +256,5 @@ class PyRunner(Runner):
 
             last = exec_plan.execution_ops[-1].logical_ops[-1]
             final_result = partition_intermediate_results[last.id()]
-            pset_id = self._part_set_cache.put_partition_set(final_result)
-            return pset_id
+            pset_entry = self._part_set_cache.put_partition_set(final_result)
+            return pset_entry
