@@ -1,7 +1,7 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Any, Callable, ClassVar, List, cast
+from typing import TYPE_CHECKING, Any, Callable, ClassVar
 
 import pyarrow as pa
 import ray
@@ -13,9 +13,12 @@ from daft.execution.logical_op_runners import (
     LogicalPartitionOpRunner,
     ReduceType,
 )
+from daft.expressions import ColumnExpression
+from daft.filesystem import glob_path
 from daft.internal.rule_runner import FixedPointPolicy, Once, RuleBatch, RuleRunner
-from daft.logical.logical_plan import LogicalPlan
+from daft.logical import logical_plan
 from daft.logical.optimizer import (
+    DropProjections,
     DropRepartition,
     FoldProjections,
     PruneColumns,
@@ -23,12 +26,14 @@ from daft.logical.optimizer import (
     PushDownLimit,
     PushDownPredicates,
 )
+from daft.logical.schema import ExpressionList
 from daft.resource_request import ResourceRequest
 from daft.runners.blocks import ArrowDataBlock, zip_blocks_as_py
 from daft.runners.partitioning import (
     PartID,
     PartitionCacheEntry,
     PartitionSet,
+    PartitionSetFactory,
     vPartition,
 )
 from daft.runners.profiler import profiler
@@ -42,11 +47,34 @@ from daft.runners.shuffle_ops import (
     Shuffler,
     SortOp,
 )
+from daft.types import ExpressionType
 
 if TYPE_CHECKING:
     from ray.data.block import Block as RayDatasetBlock
     from ray.data.dataset import Dataset as RayDataset
 
+_RAY_FROM_ARROW_REFS_AVAILABLE = True
+try:
+    from ray.data import from_arrow_refs
+except ImportError:
+    _RAY_FROM_ARROW_REFS_AVAILABLE = False
+
+
+@ray.remote
+def _glob_path_into_vpartitions(path: str, schema: ExpressionList) -> list[tuple[PartID, vPartition]]:
+    assert len(schema) == 1
+    filepath_expr = list(schema)[0]
+    filepaths = glob_path(path)
+
+    # Hardcoded to 1 filepath per partition
+    partition_refs = []
+    for i, filepath in enumerate(filepaths):
+        partition = vPartition.from_pydict({filepath_expr.name(): [filepath]}, schema=schema, partition_id=i)
+        partition_ref = ray.put(partition)
+        partition_refs.append((i, partition_ref))
+
+    return partition_refs
+
 
 @ray.remote
 def _make_ray_block_from_vpartition(partition: vPartition) -> RayDatasetBlock:
@@ -65,14 +93,21 @@ def _make_ray_block_from_vpartition(partition: vPartition) -> RayDatasetBlock:
 class RayPartitionSet(PartitionSet[ray.ObjectRef]):
     _partitions: dict[PartID, ray.ObjectRef]
 
-    def _get_all_vpartitions(self) -> list[vPartition]:
-        partition_ids = sorted(list(self._partitions.keys()))
-        assert partition_ids[0] == 0
-        assert partition_ids[-1] + 1 == len(partition_ids)
-        return cast(List[vPartition], ray.get([self._partitions[pid] for pid in partition_ids]))
+    def items(self) -> list[tuple[PartID, ray.ObjectRef]]:
+        return sorted(self._partitions.items())
+
+    def _get_merged_vpartition(self) -> vPartition:
+        ids_and_partitions = self.items()
+        assert ids_and_partitions[0][0] == 0
+        assert ids_and_partitions[-1][0] + 1 == len(ids_and_partitions)
+        all_partitions = ray.get([part for id, part in ids_and_partitions])
+        return vPartition.merge_partitions(all_partitions, verify_partition_id=False)
 
     def to_ray_dataset(self) -> RayDataset:
-        from ray.data import from_arrow_refs
+        if not _RAY_FROM_ARROW_REFS_AVAILABLE:
+            raise ImportError(
+                "Unable to import `ray.data.from_arrow_refs`. Please ensure that you have a compatible version of Ray >= 1.10 installed."
+            )
 
         blocks = [_make_ray_block_from_vpartition.remote(self._partitions[k]) for k in self._partitions.keys()]
         # NOTE: although the Ray method is called `from_arrow_refs`, this method works also when the blocks are List[T] types
@@ -111,6 +146,16 @@ class RayPartitionSet(PartitionSet[ray.ObjectRef]):
         ray.wait([o for o in self._partitions.values()])
 
 
+class RayPartitionSetFactory(PartitionSetFactory[ray.ObjectRef]):
+    def glob_filepaths(
+        self,
+        source_path: str,
+    ) -> tuple[RayPartitionSet, ExpressionList]:
+        schema = ExpressionList([ColumnExpression(self.FILEPATH_COLUMN_NAME, ExpressionType.string())]).resolve()
+        partition_refs = ray.get(_glob_path_into_vpartitions.remote(source_path, schema))
+        return RayPartitionSet({part_id: part for part_id, part in partition_refs}), schema
+
+
 class RayRunnerSimpleShuffler(Shuffler):
     def run(self, input: PartitionSet, num_target_partitions: int) -> PartitionSet:
         map_args = self._map_args if self._map_args is not None else {}
@@ -180,7 +225,7 @@ def _ray_partition_single_part_runner(
     *input_parts: vPartition,
     op_runner: RayLogicalPartitionOpRunner,
     input_node_ids: list[int],
-    nodes: list[LogicalPlan],
+    nodes: list[logical_plan.LogicalPlan],
     partition_id: int,
 ) -> vPartition:
     input_partitions = {id: val for id, val in zip(input_node_ids, input_parts)}
@@ -189,10 +234,14 @@ def _ray_partition_single_part_runner(
 
 def _get_ray_task_options(resource_request: ResourceRequest) -> dict[str, Any]:
     options = {}
+    # FYI: Ray's default resource behaviour is documented here:
+    # https://docs.ray.io/en/latest/ray-core/tasks/resources.html
     if resource_request.num_cpus is not None:
         options["num_cpus"] = resource_request.num_cpus
     if resource_request.num_gpus is not None:
         options["num_gpus"] = resource_request.num_gpus
+    if resource_request.memory_bytes is not None:
+        options["memory"] = resource_request.memory_bytes
     return options
 
 
@@ -200,7 +249,7 @@ class RayLogicalPartitionOpRunner(LogicalPartitionOpRunner):
     def run_node_list(
         self,
         inputs: dict[int, PartitionSet],
-        nodes: list[LogicalPlan],
+        nodes: list[logical_plan.LogicalPlan],
         num_partitions: int,
         resource_request: ResourceRequest,
     ) -> PartitionSet:
@@ -261,7 +310,7 @@ class RayRunner(Runner):
                 RuleBatch(
                     "PushDownLimitsAndRepartitions",
                     FixedPointPolicy(3),
-                    [PushDownLimit(), DropRepartition()],
+                    [PushDownLimit(), DropRepartition(), DropProjections()],
                 ),
             ]
         )
@@ -272,10 +321,13 @@ class RayRunner(Runner):
 
         return self._part_set_cache.put_partition_set(pset=pset)
 
-    def optimize(self, plan: LogicalPlan) -> LogicalPlan:
+    def optimize(self, plan: logical_plan.LogicalPlan) -> logical_plan.LogicalPlan:
         return self._optimizer.optimize(plan)
 
-    def run(self, plan: LogicalPlan) -> PartitionCacheEntry:
+    def partition_set_factory(self) -> PartitionSetFactory:
+        return RayPartitionSetFactory()
+
+    def run(self, plan: logical_plan.LogicalPlan) -> PartitionCacheEntry:
         plan = self.optimize(plan)
         exec_plan = ExecutionPlan.plan_from_logical(plan)
         result_partition_set: PartitionSet
@@ -303,5 +355,5 @@ class RayRunner(Runner):
 
             last = exec_plan.execution_ops[-1].logical_ops[-1]
             final_result = partition_intermediate_results[last.id()]
-            pset_id = self._part_set_cache.put_partition_set(final_result)
-            return pset_id
+            pset_entry = self._part_set_cache.put_partition_set(final_result)
+            return pset_entry
