@@ -10,6 +10,12 @@ from daft.expressions import UdfExpression
 from daft.resource_request import ResourceRequest
 from daft.runners.blocks import DataBlock
 
+_POLARS_AVAILABLE = True
+try:
+    import polars  # noqa: F401
+except ImportError:
+    _POLARS_AVAILABLE = False
+
 StatefulUDF = type  # stateful UDFs are provided as Python Classes
 StatelessUDF = Callable[..., Sequence]
 UDF = Union[StatefulUDF, StatelessUDF]
@@ -31,8 +37,9 @@ def udf(
     f: Callable | None = None,
     *,
     return_type: type,
-    num_gpus: int | None = None,
-    num_cpus: int | None = None,
+    num_gpus: int | float | None = None,
+    num_cpus: int | float | None = None,
+    memory_bytes: int | float | None = None,
 ) -> Callable:
     """Decorator for creating a UDF
 
@@ -76,6 +83,7 @@ def udf(
         return_type: The return type of the UDF
         num_gpus: How many GPUs the UDF requires for execution, used for resource allocation when running in a distributed setting
         num_cpus: How many CPUs the UDF requires for execution, used for resource allocation when running in a distributed setting
+        memory_bytes: How many bytes of memory this UDF requires for execution, used for resource allocation when running in a distributed setting
     """
     warnings.warn(
         "DEPRECATION WARNING: @udf will be deprecated in 0.1.0 in favor of @polars_udf which is much more efficient and handles Null/NaN semantics correctly"
@@ -110,7 +118,7 @@ def udf(
                 func_ret_type=func_ret_type,
                 func_args=args,
                 func_kwargs=kwargs,
-                resource_request=ResourceRequest(num_cpus=num_cpus, num_gpus=num_gpus),
+                resource_request=ResourceRequest(num_cpus=num_cpus, num_gpus=num_gpus, memory_bytes=memory_bytes),
             )
             return out_expr
 
@@ -125,8 +133,9 @@ def polars_udf(
     f: Callable | None = None,
     *,
     return_type: type,
-    num_gpus: int | None = None,
-    num_cpus: int | None = None,
+    num_gpus: int | float | None = None,
+    num_cpus: int | float | None = None,
+    memory_bytes: int | float | None = None,
 ) -> Callable:
     """Decorator for creating a Polars UDF
 
@@ -168,7 +177,11 @@ def polars_udf(
         return_type: The return type of the UDF
         num_gpus: How many GPUs the UDF requires for execution, used for resource allocation when running in a distributed setting
         num_cpus: How many CPUs the UDF requires for execution, used for resource allocation when running in a distributed setting
+        memory_bytes: How many bytes of memory this UDF requires for execution, used for resource allocation when running in a distributed setting
     """
+    if not _POLARS_AVAILABLE:
+        raise ImportError("polars_udf requires polars to be installed")
+
     func_ret_type = ExpressionType.from_py_type(return_type)
 
     def udf_decorator(func: UDF) -> Callable:
@@ -198,7 +211,7 @@ def polars_udf(
                 func_ret_type=func_ret_type,
                 func_args=args,
                 func_kwargs=kwargs,
-                resource_request=ResourceRequest(num_cpus=num_cpus, num_gpus=num_gpus),
+                resource_request=ResourceRequest(num_cpus=num_cpus, num_gpus=num_gpus, memory_bytes=memory_bytes),
             )
             return out_expr
 
